diff mbox series

[committed] libstdc++: Fix _OutputIteratorConcept checks in algorithms

Message ID YVNsIQD8uACFtS3x@redhat.com
State New
Headers show
Series [committed] libstdc++: Fix _OutputIteratorConcept checks in algorithms | expand

Commit Message

Jonathan Wakely Sept. 28, 2021, 7:25 p.m. UTC
The _OutputIteratorConcept should be checked using the correct value
category. The std::move_backward and std::copy_backward algorithms
should use _OutputIteratorConcept instead of _ConvertibleConcept.

In order to use the correct value category, the concept should use a
function that returns _ValueT instead of using an lvalue data member.

Signed-off-by: Jonathan Wakely <jwakely@redhat.com>

libstdc++-v3/ChangeLog:

	* include/bits/boost_concept_check.h (_OutputIteratorConcept):
	Use a function to preserve value category of the type.
	* include/bits/stl_algobase.h (copy, move, fill_n): Use a
	reference as the second argument for _OutputIteratorConcept.
	(copy_backward, move_backward): Use _OutputIteratorConcept
	instead of _ConvertibleConcept.

Tested x86_64-linux. Committed to trunk.
commit 45a8cd256934be3770f7e000db7b13f10eabee9a
Author: Jonathan Wakely <jwakely@redhat.com>
Date:   Fri Sep 24 15:35:20 2021

    libstdc++: Fix _OutputIteratorConcept checks in algorithms
    
    The _OutputIteratorConcept should be checked using the correct value
    category. The std::move_backward and std::copy_backward algorithms
    should use _OutputIteratorConcept instead of _ConvertibleConcept.
    
    In order to use the correct value category, the concept should use a
    function that returns _ValueT instead of using an lvalue data member.
    
    Signed-off-by: Jonathan Wakely <jwakely@redhat.com>
    
    libstdc++-v3/ChangeLog:
    
            * include/bits/boost_concept_check.h (_OutputIteratorConcept):
            Use a function to preserve value category of the type.
            * include/bits/stl_algobase.h (copy, move, fill_n): Use a
            reference as the second argument for _OutputIteratorConcept.
            (copy_backward, move_backward): Use _OutputIteratorConcept
            instead of _ConvertibleConcept.
diff --git a/libstdc++-v3/include/bits/stl_algobase.h b/libstdc++-v3/include/bits/stl_algobase.h
index d0c49628d7f..e1443b8a92a 100644
--- a/libstdc++-v3/include/bits/stl_algobase.h
+++ b/libstdc++-v3/include/bits/stl_algobase.h
@@ -613,7 +613,7 @@  _GLIBCXX_END_NAMESPACE_CONTAINER
       // concept requirements
       __glibcxx_function_requires(_InputIteratorConcept<_II>)
       __glibcxx_function_requires(_OutputIteratorConcept<_OI,
-	    typename iterator_traits<_II>::value_type>)
+	    typename iterator_traits<_II>::reference>)
       __glibcxx_requires_can_increment_range(__first, __last, __result);
 
       return std::__copy_move_a<__is_move_iterator<_II>::__value>
@@ -646,7 +646,7 @@  _GLIBCXX_END_NAMESPACE_CONTAINER
       // concept requirements
       __glibcxx_function_requires(_InputIteratorConcept<_II>)
       __glibcxx_function_requires(_OutputIteratorConcept<_OI,
-	    typename iterator_traits<_II>::value_type>)
+	    typename iterator_traits<_II>::value_type&&>)
       __glibcxx_requires_can_increment_range(__first, __last, __result);
 
       return std::__copy_move_a<true>(std::__miter_base(__first),
@@ -850,9 +850,8 @@  _GLIBCXX_END_NAMESPACE_CONTAINER
       // concept requirements
       __glibcxx_function_requires(_BidirectionalIteratorConcept<_BI1>)
       __glibcxx_function_requires(_Mutable_BidirectionalIteratorConcept<_BI2>)
-      __glibcxx_function_requires(_ConvertibleConcept<
-	    typename iterator_traits<_BI1>::value_type,
-	    typename iterator_traits<_BI2>::value_type>)
+      __glibcxx_function_requires(_OutputIteratorConcept<_BI2,
+	    typename iterator_traits<_BI1>::reference>)
       __glibcxx_requires_can_decrement_range(__first, __last, __result);
 
       return std::__copy_move_backward_a<__is_move_iterator<_BI1>::__value>
@@ -886,9 +885,8 @@  _GLIBCXX_END_NAMESPACE_CONTAINER
       // concept requirements
       __glibcxx_function_requires(_BidirectionalIteratorConcept<_BI1>)
       __glibcxx_function_requires(_Mutable_BidirectionalIteratorConcept<_BI2>)
-      __glibcxx_function_requires(_ConvertibleConcept<
-	    typename iterator_traits<_BI1>::value_type,
-	    typename iterator_traits<_BI2>::value_type>)
+      __glibcxx_function_requires(_OutputIteratorConcept<_BI2,
+	    typename iterator_traits<_BI1>::value_type&&>)
       __glibcxx_requires_can_decrement_range(__first, __last, __result);
 
       return std::__copy_move_backward_a<true>(std::__miter_base(__first),
@@ -1144,7 +1142,7 @@  _GLIBCXX_END_NAMESPACE_CONTAINER
     fill_n(_OI __first, _Size __n, const _Tp& __value)
     {
       // concept requirements
-      __glibcxx_function_requires(_OutputIteratorConcept<_OI, _Tp>)
+      __glibcxx_function_requires(_OutputIteratorConcept<_OI, const _Tp&>)
 
       return std::__fill_n_a(__first, std::__size_to_integer(__n), __value,
 			       std::__iterator_category(__first));
diff mbox series

Patch

diff --git a/libstdc++-v3/include/bits/boost_concept_check.h b/libstdc++-v3/include/bits/boost_concept_check.h
index 5c87e32f36b..ba36c24abec 100644
--- a/libstdc++-v3/include/bits/boost_concept_check.h
+++ b/libstdc++-v3/include/bits/boost_concept_check.h
@@ -464,10 +464,10 @@  struct _Aux_require_same<_Tp,_Tp> { typedef _Tp _Type; };
       __function_requires< _AssignableConcept<_Tp> >();
       ++__i;                            // require preincrement operator
       __i++;                            // require postincrement operator
       *__i++ = __t;                     // require postincrement and assignment
     }
     _Tp __i;
-    _ValueT __t;
+    _ValueT __val() const;
   };
 
   template <class _Tp>
@@ -464,10 +464,10 @@  struct _Aux_require_same<_Tp,_Tp> { typedef _Tp _Type; };
       ++__i;                            // require preincrement operator
       __i++;                            // require postincrement operator
-      *__i++ = __t;                     // require postincrement and assignment
+      *__i++ = __val();                 // require postincrement and assignment
     }
     _Tp __i;